"""Models for Gaggiuino"""

from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Literal


@dataclass(frozen=True)
class GaggiuinoShotDataPoints:
    pressure: list[int] | None = None
    pumpFlow: list[int] | None = None
    shotWeight: list[int] | None = None
    targetPressure: list[int] | None = None
    targetPumpFlow: list[int] | None = None
    targetTemperature: list[int] | None = None
    temperature: list[int] | None = None
    timeInShot: list[int] | None = None
    waterPumped: list[int] | None = None
    weightFlow: list[int] | None = None


@dataclass(frozen=True)
class GaggiuinoProfilePhaseStopCondition:
    """
    'stopConditions': {
        'pressureAbove': 2,
        'time': 15000,
        'weight': 0.1
    },
    """

    pressureAbove: int | None = None
    time: int | None = None
    weight: float | None = None


@dataclass(frozen=True)
class GaggiuinoProfilePhaseTarget:
    """
    'target': {
        'curve': 'INSTANT',
        'end': 2,
        'time': 10000
    },
    """

    curve: str
    end: int
    time: int


@dataclass(frozen=True)
class GaggiuinoProfileType:
    """
    'type': 'FLOW'
    """

    type: Literal['FLOW', 'PRESSURE']


@dataclass(frozen=True)
class GaggiuinoProfilePhase:
    """
    {
        'restriction': 2,
        'skip': False,
        'stopConditions': {
            'pressureAbove': 2,
            'time': 15000,
            'weight': 0.1
        },
        'target': {
            'curve': 'INSTANT',
            'end': 2,
            'time': 10000
        },
        'type': 'FLOW'
    },
    """

    restriction: int
    skip: bool
    stopConditions: GaggiuinoProfilePhaseStopCondition
    type: GaggiuinoProfileType


@dataclass(frozen=True)
class GaggiuinoProfile:
    """
    'profile': {
        'globalStopConditions': {
            'weight': 50
        },
        'id': 8,
        'name': '_Long',
        'phases': [
            {
                'restriction': 2,
                'skip': False,
                'stopConditions': {
                    'pressureAbove': 2,
                    'time': 15000,
                    'weight': 0.1
                },
                'target': {
                    'curve': 'INSTANT',
                    'end': 2,
                    'time': 10000
                },
                'type': 'FLOW'
            },
            {
                'restriction': 1,
                'skip': False,
                'stopConditions': {
                    'time': 15000
                },
                'target': {
                    'curve': 'INSTANT',
                    'end': 0
                },
                'type': 'FLOW'
            }, {
                'restriction': 9,
                'skip': False,
                'stopConditions': {},
                'target': {
                    'curve': 'EASE_IN_OUT',
                    'end': 1.5,
                    'start': 2,
                    'time': 15000
                },
                'type': 'FLOW'
            }
        ],
        'recipe': {},
        'waterTemperature': 90
    },
    """

    id: int
    name: str
    selected: bool | None = None
    globalStopConditions: dict[str, Any] | None = None
    phases: list[GaggiuinoProfilePhase] | None = None
    recipe: dict[str, Any] | None = None
    waterTemperature: int | None = None


@dataclass(frozen=True)
class GaggiuinoShot:
    """
    {
        'datapoints': {
            'pressure': [
                3, 3, 3, ...
            ],
            'pumpFlow': [
                0, 6, 12, 12, ...
            ],
            'shotWeight': [
                0, 0, 0, ...
            ],
            'targetPressure': [
                20, 20, 20, ...
            ],
            'targetPumpFlow': [
                20, 20, 20, ...
            ],
            'targetTemperature': [
                900, 900, 900, ...
            ],
            'temperature': [
                898, 898, 898, ...
            ],
            'timeInShot': [
                2, 3, 5, ...
            ],
            'waterPumped': [
                0, 2, 4, ...
            ],
            'weightFlow': [
                0, 0, 0, ...
            ]
        },
        'duration': 648,
        'id': 1,
        'profile': {
            'globalStopConditions': {
                'weight': 50
            },
            'id': 8,
            'name': '_Long',
            'phases': [
                {
                    'restriction': 2,
                    'skip': False,
                    'stopConditions': {
                        'pressureAbove': 2,
                        'time': 15000,
                        'weight': 0.1
                    },
                    'target': {
                        'curve': 'INSTANT',
                        'end': 2,
                        'time': 10000
                    },
                    'type': 'FLOW'
                },
                {
                    'restriction': 1,
                    'skip': False,
                    'stopConditions': {
                        'time': 15000
                    },
                    'target': {
                        'curve': 'INSTANT',
                        'end': 0
                    },
                    'type': 'FLOW'
                }, {
                    'restriction': 9,
                    'skip': False,
                    'stopConditions': {},
                    'target': {
                        'curve': 'EASE_IN_OUT',
                        'end': 1.5,
                        'start': 2,
                        'time': 15000
                    },
                    'type': 'FLOW'
                }
            ],
            'recipe': {},
            'waterTemperature': 90
        },
        'timestamp': 1731316192
    }
    """

    datapoints: GaggiuinoShotDataPoints
    duration: int
    id: int
    profile: GaggiuinoProfile
    timestamp: int


@dataclass
class GaggiuinoStatus:
    """
    {
      "upTime": "89107",
      "profileId": "7",
      "profileName": "OFF",
      "targetTemperature": "15.000000",
      "temperature": "22.500000",
      "pressure": "-0.028054",
      "waterLevel": "100",
      "weight": "0.000000",
      "brewSwitchState": "false",
      "steamSwitchState": "false"
    }

    """

    upTime: int
    profileId: int
    profileName: str
    targetTemperature: float
    temperature: float
    temperature: float
    pressure: float
    waterLevel: int
    weight: float
    brewSwitchState: bool
    steamSwitchState: bool


@dataclass(frozen=True)
class GaggiuinoLatestShotResult:
    """
    [
        {
            "lastShotId": "100"
        }
    ]
    """

    lastShotId: int



"""Exceptions for gaggiuino"""


class GaggiuinoError(Exception):
    """Generic Gaggiuino exception."""


class GaggiuinoConnectionError(GaggiuinoError):
    """Gaggiuino connection error exception."""


class GaggiuinoEndpointNotFoundError(GaggiuinoError):
    """Gaggiuino endpoint not found exception."""



"""End-to-end tests for Gaggiuino API."""

import os
import pytest
import pytest_asyncio
from gaggiuino_api import (
    GaggiuinoAPI,
    GaggiuinoEndpointNotFoundError,
    GaggiuinoProfile,
    GaggiuinoShot,
    GaggiuinoStatus,
)
from gaggiuino_api.models import GaggiuinoLatestShotResult
from gaggiuino_api.const import DEFAULT_BASE_URL

pytest_plugins = ('pytest_asyncio',)


@pytest_asyncio.fixture(loop_scope="session")
async def api_client():
    """Fixture to provide an API client."""
    async with GaggiuinoAPI(base_url=DEFAULT_BASE_URL) as gapi:
        yield gapi


@pytest.mark.asyncio(loop_scope="session")
async def test_api_connection(api_client):
    """Test basic API connection."""
    assert api_client is not None


@pytest.mark.asyncio(loop_scope="session")
async def test_get_profiles(api_client):
    """Test getting profiles from the API."""
    profiles = await api_client.get_profiles()
    assert isinstance(profiles, list)
    if profiles:  # If there are any profiles
        assert isinstance(profiles[0], GaggiuinoProfile)
        assert all(isinstance(profile, GaggiuinoProfile) for profile in profiles)
        assert all(hasattr(profile, 'id') for profile in profiles)
        assert all(getattr(profile, 'id', None) is not None for profile in profiles)
        assert all(hasattr(profile, 'name') for profile in profiles)
        assert all(hasattr(profile, 'phases') for profile in profiles)


@pytest.mark.asyncio(loop_scope="session")
async def test_select_profile(api_client):
    """Test selecting a profile."""
    profiles = await api_client.get_profiles()
    if not profiles:
        pytest.skip("No profiles available for testing")

    profile_off_name = os.getenv('GAGGIUINO_PROFILE_OFF', 'OFF')
    profile_test_name = os.getenv('GAGGIUINO_PROFILE_TEST', 'test')

    profile_off = next((_ for _ in profiles if _.name == profile_off_name), None)
    assert profile_off, "Cannot check withoout an OFF profile"
    profile_test = next((_ for _ in profiles if _.name == profile_test_name), None)
    assert profile_off, "Cannot check withoout a test profile"

    # Test with profile object
    result = await api_client.select_profile(profile_test)
    assert result is True

    # Test with profile ID
    result = await api_client.select_profile(profile_off.id)
    assert result is True


@pytest.mark.asyncio(loop_scope="session")
async def test_invalid_profile_selection(api_client):
    """Test selecting an invalid profile ID."""
    with pytest.raises(
        GaggiuinoEndpointNotFoundError
    ):  # Will raise when profile not found
        await api_client.select_profile(99999)  # Using an arbitrary large ID


@pytest.mark.asyncio(loop_scope="session")
async def test_invalid_shot_request(api_client):
    """Test requesting an invalid shot ID."""
    with pytest.raises(GaggiuinoEndpointNotFoundError):
        await api_client.get_shot(99999)  # Using an arbitrary large ID


@pytest.mark.asyncio(loop_scope="session")
async def test_get_status(api_client):
    """Test getting profiles from the API."""
    status = await api_client.get_status()
    assert isinstance(status, GaggiuinoStatus)


@pytest.mark.asyncio(loop_scope="session")
async def test_get_shot(api_client):
    """Test getting the latst shot id."""
    latest_shot_id = await api_client.get_latest_shot_id()
    assert isinstance(latest_shot_id, GaggiuinoLatestShotResult)
    latest_shot_id = latest_shot_id.lastShotId

    shot_data = await api_client.get_shot(latest_shot_id)
    if shot_data is None:
        pytest.skip("No shot data available for testing")

    assert isinstance(shot_data, GaggiuinoShot)



from .api import GaggiuinoAPI
from .exceptions import (
    GaggiuinoError,
    GaggiuinoConnectionError,
    GaggiuinoEndpointNotFoundError,
)
from .models import (
    GaggiuinoShot,
    GaggiuinoShotDataPoints,
    GaggiuinoProfile,
    GaggiuinoProfileType,
    GaggiuinoProfilePhase,
    GaggiuinoProfilePhaseTarget,
    GaggiuinoProfilePhaseStopCondition,
    GaggiuinoStatus,
)

__all__ = [
    'GaggiuinoAPI',
    'GaggiuinoError',
    'GaggiuinoConnectionError',
    'GaggiuinoEndpointNotFoundError',
    'GaggiuinoShot',
    'GaggiuinoShotDataPoints',
    'GaggiuinoProfile',
    'GaggiuinoProfileType',
    'GaggiuinoProfilePhase',
    'GaggiuinoProfilePhaseTarget',
    'GaggiuinoProfilePhaseStopCondition',
    'GaggiuinoStatus',
]



"""Gaggiuino API Wrapper."""

from __future__ import annotations
import sys

import asyncio
import logging
from typing import Type, Any, Literal
from urllib import parse as urllib_parse

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectionError

from gaggiuino_api.const import DEFAULT_BASE_URL
from gaggiuino_api.exceptions import (
    GaggiuinoError,
    GaggiuinoConnectionError,
    GaggiuinoEndpointNotFoundError,
)
from gaggiuino_api.models import (
    GaggiuinoProfile,
    GaggiuinoShot,
    GaggiuinoStatus,
    GaggiuinoLatestShotResult,
)

if sys.platform == 'win32':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

_LOGGER = logging.getLogger(__name__)


class GaggiuinoClient:
    """Initialise a client to receive Server Sent Events (SSE)"""

    def __init__(self, base_url: str = DEFAULT_BASE_URL, session: ClientSession = None):
        self.session = session
        self.base_url = base_url
        self.headers = {}
        self.post_headers = {"Content-Type": "application/x-www-form-urlencoded"}
        self.close_session = False
        self.timeout = 15

    async def __aenter__(self) -> "GaggiuinoClient":
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc: BaseException | None,
        tb: object | None,
    ) -> None:
        await self.disconnect()

    async def connect(self) -> None:
        """Open the session"""
        if self.session is None:
            self.close_session = True
            self.session = ClientSession(headers=self.headers)

    async def disconnect(self) -> None:
        """Close the session if it was created internally"""
        if self.session is not None and self.close_session:
            await self.session.close()
            self.session = None
            self.close_session = False

    async def post(self, url: str, params: dict = None) -> bool:
        assert self.session is not None, "Session not created"

        data = urllib_parse.urlencode(params) if params is not None else None

        try:
            async with self.session.post(
                url,
                data=data,
                headers=self.post_headers,
            ) as response:
                if response.status == 404:
                    raise GaggiuinoEndpointNotFoundError("endpoint not found")
                await response.text(encoding="utf-8")
                return response.status == 200
        except ClientConnectionError as err:
            raise GaggiuinoConnectionError("Connection failed") from err
        except GaggiuinoEndpointNotFoundError as err:
            raise GaggiuinoEndpointNotFoundError from err
        except Exception as err:
            raise GaggiuinoError("Unhandled exception") from err

    async def get(
        self,
        url: str | None = None,
        params: dict[str, Any] = None,
    ) -> Any:
        assert self.session is not None, "Session not created"

        params = params or {}
        url = url or self.base_url

        try:
            async with self.session.get(
                url,
                headers=self.headers,
                params=params,
                timeout=self.timeout,
            ) as response:
                if response.status == 404:
                    raise GaggiuinoEndpointNotFoundError("endpoint not found")

                return await response.json()
        except ClientConnectionError as err:
            raise GaggiuinoConnectionError("Connection failed") from err
        except GaggiuinoEndpointNotFoundError as err:
            raise GaggiuinoEndpointNotFoundError from err
        except Exception as err:
            raise GaggiuinoError("Unhandled exception") from err


class GaggiuinoAPI(GaggiuinoClient):
    def __init__(
        self,
        base_url: str = DEFAULT_BASE_URL,
        *,
        session: ClientSession | None = None,
    ) -> None:
        super().__init__(base_url=base_url, session=session)
        self.api_base = f"{self.base_url}/api"
        self._profile: GaggiuinoProfile | None = None
        self._profiles: list[GaggiuinoProfile] | None = None
        self._status: GaggiuinoStatus | None = None

    @property
    def profile(self):
        if self._status is None:
            _LOGGER.debug(
                "Cannot get the currently selected profile. Use get_status() first."
            )
            return None

        self._profile = GaggiuinoProfile(
            id=self._status.profileId, name=self._status.profileName
        )
        if self._profiles:
            for profile in self._profiles:
                if profile.id == self._status.profileId:
                    self._profile = profile
                    break

        _LOGGER.debug("Current profile: %s", self._profile)
        return self._profile

    @profile.setter
    async def profile(self, profile: GaggiuinoProfile | int):
        if await self._select_profile(profile):
            self._profile = profile

    async def get_profiles(self) -> list[GaggiuinoProfile] | None:
        url = f"{self.api_base}/profiles/all"
        try:
            profiles: list[dict[str, Any]] = await self.get(url)
        except GaggiuinoEndpointNotFoundError as err:
            raise GaggiuinoEndpointNotFoundError("Profile not found") from err
        except Exception as err:
            raise GaggiuinoError("Unhandled exception") from err
        if profiles is None:
            return None

        self._profiles = [GaggiuinoProfile(**_) for _ in profiles]
        return self._profiles

    async def _select_profile(self, profile_id: int) -> bool:
        url = f"{self.api_base}/profile-select/{profile_id}"
        try:
            return await self.post(url)
        except GaggiuinoEndpointNotFoundError as err:
            raise GaggiuinoEndpointNotFoundError("Profile not found") from err
        except Exception as err:
            raise GaggiuinoError("Unhandled exception") from err

    async def select_profile(self, profile: GaggiuinoProfile | int) -> bool:
        profile_id = profile
        if isinstance(profile, GaggiuinoProfile):
            profile_id = profile.id

        return await self._select_profile(profile_id=profile_id)

    async def _get_shot(self, shot_id: int | Literal["latest"]) -> dict:
        url = f"{self.api_base}/shots/{shot_id}"
        try:
            shot = await self.get(url)
        except GaggiuinoEndpointNotFoundError as err:
            raise GaggiuinoEndpointNotFoundError("Shot not found") from err
        except Exception as err:
            raise GaggiuinoError("Unhandled exception") from err

        return shot

    async def get_shot(self, shot_id: int) -> GaggiuinoShot | None:
        shot = await self._get_shot(shot_id)
        if shot is None:
            _LOGGER.debug("Couldn't retrieve shot %s", shot_id)
            return None

        return GaggiuinoShot(**shot)

    async def get_status(self) -> GaggiuinoStatus | None:
        url = f"{self.api_base}/system/status"
        try:
            status: list[dict[str, Any]] = await self.get(url)
        except GaggiuinoEndpointNotFoundError as err:
            raise GaggiuinoEndpointNotFoundError("Shot not found") from err
        except Exception as err:
            raise GaggiuinoError("Unhandled exception") from err

        if status:
            self._status = GaggiuinoStatus(**status[0])
            return self._status

    async def get_latest_shot_id(self):
        latest_shots = await self._get_shot("latest")
        if latest_shots is None:
            _LOGGER.debug("Couldn't retrieve the latest shot")
            return None

        return GaggiuinoLatestShotResult(**latest_shots[0])


async def _main():
    async with GaggiuinoAPI() as gapi:
        _status = await gapi.get_status()
        _profiles = await gapi.get_profiles()
        _latest_shot_id_result = await gapi.get_latest_shot_id()
        _latest_shot_id = _latest_shot_id_result.lastShotId
        _shot = await gapi.get_shot(_latest_shot_id)
    pass


if __name__ == '__main__':
    asyncio.run(_main())
